﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;

/* SunVoxSharp - C# binding for libSunVox
 * (c) 2012 ed <irc.rizon.net>
 * 
 * Protected by same license as libSunVox itself: source code
 * can be distributed freely, with or without changes, as long
 * as this block-comment is unmodified and present in its entirety.
 * 
 * You are free to distribute this as a binary, with or without
 * changes, as long as this text is displayed in the program's
 * about-screen or staffroll in its entirety:
 * 
 * Sound powered by SunVox modular synthesizer
 * Copyright (c) 2002 - 2012, Alexander Zolotov
 *      nightradio@gmail.com, WarmPlace.ru
 * 
 * Ogg Vorbis 'Tremor' integer playback codec
 * Copyright (c) 2002, Xiph.org Foundation
 */

namespace SunVoxSharp
{
    public class SunVoxCore
    {
        SunVoxTrack rendertrack;
        [DllImport("sunvox.dll")] static extern int sv_init(string dev, int freq, int channels, int flags);
        [DllImport("sunvox.dll")] static extern int sv_audio_callback(byte[] audio_buf, int frames, int latency, int out_time);
        [DllImport("sunvox.dll")] static extern int sv_deinit();

        //sv_send_event() parameters:
        //  handle;
        //  channel_num: from 0 to 7;
        //  note: 0 - nothing; 1..127 - note num; 128 - note off; 129, 130... - see NOTECMD_xxx defines;
        //  vel: velocity 1..129; 0 - default;
        //  module: 0 - nothing; 1..255 - module number;
        //  ctl: CCXX. CC - number of controller. XX - std effect;
        //  ctl_val: value of controller.

        // s/^#define ([^\s]*)\s+(.*)/public static int $1 = $2/
        public static int SV_INIT_FLAG_NO_DEBUG_OUTPUT = (1 << 0);
        public static int SV_INIT_FLAG_USER_AUDIO_CALLBACK  = ( 1 << 1 );
        public static int SV_INIT_FLAG_AUDIO_INT16          = ( 1 << 2 );
        public static int SV_INIT_FLAG_AUDIO_FLOAT32        = ( 1 << 3 );
        public static int SV_INIT_FLAG_ONE_THREAD           = ( 1 << 4 );

        public static int SV_MODULE_FLAG_EXISTS = 1;
        public static int SV_MODULE_FLAG_EFFECT  = 2;
        public static int SV_MODULE_INPUTS_OFF   = 16;
        public static int SV_MODULE_OUTPUTS_OFF  = (16 + 8);
        public static int SV_MODULE_INPUTS_MASK  = (255 << SV_MODULE_INPUTS_OFF);
        public static int SV_MODULE_OUTPUTS_MASK = (255 << SV_MODULE_OUTPUTS_OFF);

        public static int SV_STYPE_INT16 = 0;
        public static int SV_STYPE_INT32 = 1;
        public static int SV_STYPE_FLOAT32 = 2;
        public static int SV_STYPE_FLOAT64 = 3;

        public static int NOTECMD_ALL_NOTES_OFF = 129; /* notes of all synths off */
        public static int NOTECMD_CLEAN_SYNTHS = 130; /* stop and clean all synths */
        public static int NOTECMD_STOP = 131;
        public static int NOTECMD_PLAY = 132;

        int samplerate;
        List<bool> freehandles;
        bool renderingToSoundcard;
        static bool isInitialized = false;
        List<SunVoxTrack> tracks;
        
        public int VER_MAJOR;
        public int VER_MINOR;
        public int VER_PATCH;

        /// <summary>
        /// Load the SunVox Library. Only 1 (one) SunVox Library can be loaded at any given time.
        /// If you need to load it multiple times (for instamce, to pre-render sound effects to raster files)
        /// please first create svCore object with SunVoxCore(false, dontcare), render all your SFX to wav/memory,
        /// then dispose it before re-creating it with SunVoxCore(true, dontcare) for BGM usage.
        /// </summary>
        /// <param name="useDefaultSoundcard">true if you want SunVox to handle audio output itself, false for writing to XNA/wav/memory.</param>
        /// <param name="useLowQuality">true downgrades internal audio from 32bit float to 16bit fixed, boosting speed and giving everything a lo-fi feel</param>
        public SunVoxCore(bool useDefaultSoundcard, bool useLowQuality)
        {
            if (SunVoxCore.isInitialized)
                throw new Exception("please dispose the old SunVoxCore before initializing another one");

            freehandles = new List<bool>();
            int flags = SV_INIT_FLAG_NO_DEBUG_OUTPUT;
            renderingToSoundcard = useDefaultSoundcard;
            if (!useDefaultSoundcard)
            {
                //flags += SV_INIT_FLAG_USER_AUDIO_CALLBACK;
                useLowQuality = true; //TODO: Really?
            }
            if (useLowQuality)
            {
                flags += SV_INIT_FLAG_AUDIO_INT16;
            }
            int ver = sv_init(null, 44100, 2, flags);
            VER_MAJOR = ((ver >> 16) & 255);
            VER_MINOR = ((ver >> 8)  & 255);
            VER_PATCH = ((ver)       & 255);
            SunVoxCore.isInitialized = true;
            this.samplerate = 44100; //lolhack
            tracks = new List<SunVoxTrack>();
        }
        /// <summary>
        /// create player and load song by filename
        /// </summary>
        /// <param name="file">path to file</param>
        public SunVoxTrack load(string file)
        {
            SunVoxTrack track = new SunVoxTrack(freehandles, samplerate, file);
            tracks.Add(track);
            return track;
        }
        /// <summary>
        /// create player and load song from memory
        /// </summary>
        /// <param name="data">sunvox data</param>
        public SunVoxTrack load(byte[] data)
        {
            SunVoxTrack track = new SunVoxTrack(freehandles, samplerate, data);
            tracks.Add(track);
            return track;
        }
        /// <summary>
        /// Destroy all loaded files and unload the library.
        /// Call before creating another SunVoxCore.
        /// </summary>
        public int dispose()
        {
            while (tracks.Count > 0)
            {
                tracks[0].dispose();
                tracks.RemoveAt(0);
            }
            SunVoxCore.isInitialized = false;
            return sv_deinit();
        }
        /// <summary>
        /// Render a single track (from sunvox file) to a wav-file or memory.
        /// Only works when SunVoxCore was initialized with useDefaultSoundcard = false
        /// </summary>
        /// <param name="filename">the file to render</param>
        public byte[] render(string filename, double fadeout_duration)
        {
            if (renderingToSoundcard)
                throw new Exception("SunVoxCore.render(filename) called after SunVoxCore was initialized for output to soundcard");

            rendertrack = this.load(filename);
            return render(fadeout_duration);
        }
        /// <summary>
        /// Render a single track (from memory) to a wav-file or memory.
        /// Only works when SunVoxCore was initialized with useDefaultSoundcard = false
        /// </summary>
        /// <param name="filename">the file to render</param>
        public byte[] render(byte[] sunvox_data, double fadeout_duration)
        {
            if (renderingToSoundcard)
                throw new Exception("SunVoxCore.render(sunvox_dat) called after SunVoxCore was initialized for output to soundcard");

            rendertrack = this.load(sunvox_data);
            return render(fadeout_duration);
        }

        byte[] render(double fadeout_duration)
        {
            int samples = rendertrack.totalSamples + (int)(44100 * fadeout_duration);
            double seconds = samples * 1.0 / samplerate;
            byte[] buffer = new byte[samples * 4];
            //rendertrack.play();
            rendertrack.disable_looping = true;
            rendertrack.position = -2;
            rendertrack.resume();
            while (true)
            {
                rendertrack.sendEvent(0, 130, 0, 0, 0, 0); //clean synths
                if (rendertrack.position >= -1) break;
                System.Threading.Thread.Sleep(10);
            }
            sv_audio_callback(buffer, samples, 0, 0); //(int)(rendertrack.engineTicksPerSecond * seconds));
            rendertrack.dispose();
            int first = 0;
            int last = buffer.Length;
            for (int i = 0; i < last; i++)
            {
                if (buffer[i] != 0)
                {
                    first = i;
                    break;
                }
            }
            for (int i = last - 1; i > first; i--)
            {
                if (buffer[i] != 0)
                {
                    last = i + 1;
                    break;
                }
            }
            // i love documentation
            first /= 4;
            last /= 4;
            last++;
            last *= 4;
            first *= 4;
            last -= first;
            // so feel free to add some

            last = Math.Min(buffer.Length - first, last);
            byte[] tmp = new byte[last + 0x2C];
            Array.Copy(buffer, first, tmp, 0x2C, last);
            buffer = tmp;
            byte[] header = Convert.FromBase64String("UklGRiQsqgJXQVZFZm10IBAAAAABAAIARKwAABCxAgAEABAAZGF0YQAsqgI=");
            Array.Copy(header, buffer, header.Length);
            Array.Copy(BitConverter.GetBytes((Int32)(buffer.Length - 0x08)), 0, buffer, 0x04, 4);
            Array.Copy(BitConverter.GetBytes((Int32)(buffer.Length - 0x2C)), 0, buffer, 0x28, 4);
            return buffer;
        }
    }
}
